@@ -75,7 +75,6 @@ LIBSASL_API void sasl_churn(sasl_rand_t *rpool,
 #ifdef WIN32
 LIBSASL_API int getopt(int argc, char **argv, char *optstring);
 LIBSASL_API char * getpass(const char *prompt);
-
-#endif //WIN32
+#endif /* WIN32 */
 
 #endif /* SASLUTIL_H */
