@@ -27,6 +27,8 @@ SOFTWARE.
 
 #include <sasl.h>
 #include <saslint.h>
+#include <saslutil.h>
+#include <assert.h>
 
 #ifdef HAVE_KRB
 #include <krb.h>
@@ -61,6 +63,17 @@ SOFTWARE.
 #include <security/pam_appl.h>
 #endif
 
+#ifdef HAVE_PWCHECK
+#include <errno.h>
+#include <sys/types.h>
+#include <sys/uio.h>
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+
+extern int errno;
+#endif
+
 #ifdef HAVE_KRB
 
 /* This defines the Andrew string_to_key function.  It accepts a
@@ -438,8 +451,9 @@ pam_err:
  *
  * where <secret> = MD5(<salt>, "sasldb", <pass>)
  */
-int _sasl_make_plain_secret(const char *salt, const char *passwd, int passlen,
-			    sasl_secret_t **secret)
+static int _sasl_make_plain_secret(const char *salt, 
+				   const char *passwd, int passlen,
+				   sasl_secret_t **secret)
 {
     MD5_CTX ctx;
     unsigned sec_len = 16 + 1 + 16; /* salt + "\0" + hash */
@@ -462,26 +476,84 @@ int _sasl_make_plain_secret(const char *salt, const char *passwd, int passlen,
     return SASL_OK;
 }
 
+/* returns the realm we should pretend to be in */
+static int parseuser(char **user, char **realm, const char *user_realm, 
+		     const char *serverFQDN, const char *input)
+{
+    int ret;
+    char *r;
+
+    assert(user && serverFQDN);
+
+    if (!user_realm) {
+	ret = _sasl_strdup(serverFQDN, realm, NULL);
+	if (ret == SASL_OK) {
+	    ret = _sasl_strdup(input, user, NULL);
+	}
+    } else if (user_realm[0]) {
+	ret = _sasl_strdup(user_realm, realm, NULL);
+	if (ret == SASL_OK) {
+	    ret = _sasl_strdup(input, user, NULL);
+	}
+    } else {
+	/* otherwise, we gotta get it from the user */
+	r = strchr(input, '@');
+	if (!r) {
+	    *realm = sasl_ALLOC(1);
+	    if (*realm) {
+		*realm[0] = '\0';
+		ret = SASL_OK;
+	    } else {
+		ret = SASL_NOMEM;
+	    }
+	    if (ret == SASL_OK) {
+		ret = _sasl_strdup(input, user, NULL);
+	    }
+	} else {
+	    int i;
+
+	    r++;
+	    ret = _sasl_strdup(r, realm, NULL);
+	    *user = sasl_ALLOC(r - input + 1);
+	    if (*user) {
+		for (i = 0; input[i] != '@'; i++) {
+		    *user[i] = input[i];
+		}
+		*user[i] = '\0';
+	    } else {
+		ret = SASL_NOMEM;
+	    }
+	}
+    }
+
+    return ret;
+}
+
 int _sasl_sasldb_verify_password(sasl_conn_t *conn,
-				 const char *userid, const char *passwd,
-				 const char **reply)
+				 const char *userstr, const char *passwd,
+				 const char *user_realm, const char **reply)
 {
     sasl_server_getsecret_t *getsec;
     void *context;
     int ret;
     sasl_secret_t *secret = NULL;
     sasl_secret_t *construct = NULL;
+    char *userid, *realm;
 
-    if (!userid || !passwd) {
+    if (!userstr || !passwd) {
 	return SASL_BADPARAM;
     }
     if (reply) { *reply = NULL; }
+    ret = parseuser(&userid, &realm, user_realm, conn->serverFQDN, userstr);
+    if (ret != SASL_OK) {
+	return ret;
+    }
     ret = _sasl_getcallback(conn, SASL_CB_SERVER_GETSECRET, &getsec, &context);
     if (ret != SASL_OK) {
 	return ret;
     }
 
-    ret = getsec(context, "PLAIN", userid, &secret);
+    ret = getsec(context, "PLAIN", userid, realm, &secret);
     if (ret != SASL_OK) {
 	return ret;
     }
@@ -505,3 +577,206 @@ int _sasl_sasldb_verify_password(sasl_conn_t *conn,
     return ret;
 }
 
+/* this routine sets the sasldb password given a user/pass */
+int _sasl_sasldb_set_pass(sasl_conn_t *conn,
+			  const char *userstr, 
+			  const char *pass,
+			  unsigned passlen,
+			  const char *user_realm,
+			  int flags,
+			  const char **errstr)
+{
+    char *userid, *realm;
+    int ret = SASL_OK;
+
+    if (*errstr) {
+	*errstr = NULL;
+    }
+
+    ret = parseuser(&userid, &realm, user_realm, conn->serverFQDN, userstr);
+    if (ret != SASL_OK) {
+	return ret;
+    }
+
+    if (pass != NULL && !(flags & SASL_SET_DISABLE)) {
+	/* set the password */
+	sasl_secret_t *sec = NULL;
+	char salt[16];
+	sasl_rand_t *rpool;
+	sasl_server_getsecret_t *getsec;
+	sasl_server_putsecret_t *putsec;
+	void *context;
+
+	/* if SASL_SET_CREATE is set, we don't want to overwrite an
+	   existing account */
+	if (flags & SASL_SET_CREATE) {
+	    ret = _sasl_getcallback(conn, SASL_CB_SERVER_GETSECRET,
+				    &getsec, &context);
+	    if (ret == SASL_OK) {
+		getsec(context, "PLAIN", userid, realm, &sec);
+		if (ret == SASL_OK) {
+		    sasl_free_secret(&sec);
+		    ret = SASL_NOCHANGE;
+		}
+		if (ret == SASL_NOUSER) {
+		    /* hmmm, don't want to change this */
+		    ret = SASL_OK;
+		}
+	    }
+	}
+	
+	/* ret == SASL_OK iff we still want to set this password */
+	if (ret == SASL_OK) {
+	    ret = sasl_randcreate(&rpool);
+	}
+	if (ret == SASL_OK) {
+	    sasl_rand(rpool, salt, 16);
+	    ret = _sasl_make_plain_secret(salt, pass, passlen, &sec);
+	}
+	if (ret == SASL_OK) {
+	    ret = _sasl_getcallback(conn, SASL_CB_SERVER_PUTSECRET, 
+				    &putsec, &context);
+	}
+	if (ret == SASL_OK) {
+	    ret = putsec(context, "PLAIN", userid, realm, sec);
+	}
+	if (ret != SASL_OK) {
+	    _sasl_log(conn, SASL_LOG_ERR, NULL, ret, 0,
+		      "failed to set plaintext secret for %s: %z", userid);
+	}
+	if (rpool) {
+	    sasl_randfree(&rpool);
+	}
+	if (sec) {
+	    sasl_free_secret(&sec);
+	}
+    } else {
+	sasl_server_putsecret_t *putsec;
+	void *context;
+
+	ret = _sasl_getcallback(conn, SASL_CB_SERVER_PUTSECRET, 
+				&putsec, &context);
+	if (ret == SASL_OK) {
+	    ret = putsec(context, "PLAIN", userid, realm, NULL);
+	}
+    }
+
+    sasl_FREE(realm);
+    return ret;
+}
+
+
+#ifdef HAVE_PWCHECK
+/*
+ * Keep calling the writev() system call with 'fd', 'iov', and 'iovcnt'
+ * until all the data is written out or an error occurs.
+ */
+static int retry_writev(int fd, struct iovec *iov, int iovcnt)
+{
+    int n;
+    int i;
+    int written = 0;
+    static int iov_max =
+#ifdef MAXIOV
+	MAXIOV
+#else
+#ifdef IOV_MAX
+	IOV_MAX
+#else
+	8192
+#endif
+#endif
+	;
+    
+    for (;;) {
+	while (iovcnt && iov[0].iov_len == 0) {
+	    iov++;
+	    iovcnt--;
+	}
+
+	if (!iovcnt) return written;
+
+	n = writev(fd, iov, iovcnt > iov_max ? iov_max : iovcnt);
+	if (n == -1) {
+	    if (errno == EINVAL && iov_max > 10) {
+		iov_max /= 2;
+		continue;
+	    }
+	    if (errno == EINTR) continue;
+	    return -1;
+	}
+
+	written += n;
+
+	for (i = 0; i < iovcnt; i++) {
+	    if (iov[i].iov_len > n) {
+		iov[i].iov_base = (char *)iov[i].iov_base + n;
+		iov[i].iov_len -= n;
+		break;
+	    }
+	    n -= iov[i].iov_len;
+	    iov[i].iov_len = 0;
+	}
+
+	if (i == iovcnt) return written;
+    }
+}
+
+/* pwcheck daemon-authenticated login */
+int _sasl_pwcheck_verify_password(sasl_conn_t *conn,
+				  const char *userid, const char *passwd,
+				  const char **reply)
+{
+    int s;
+    struct sockaddr_un srvaddr;
+    int r;
+    struct iovec iov[10];
+    static char response[1024];
+    int start, n;
+    char *pwpath[1024];
+    sasl_getopt_t *getopt;
+    void *context;
+
+    if (reply) { *reply = NULL; }
+
+    strcpy(pwpath, PWCHECKDIR);
+    strcat(pwpath, "/pwcheck");
+
+    s = socket(AF_UNIX, SOCK_STREAM, 0);
+    if (s == -1) return errno;
+
+    memset((char *)&srvaddr, 0, sizeof(srvaddr));
+    srvaddr.sun_family = AF_UNIX;
+    strcpy(srvaddr.sun_path, pwpath);
+    r = connect(s, (struct sockaddr *)&srvaddr, sizeof(srvaddr));
+    if (r == -1) {
+	if (reply) { *reply = "cannot connect to pwcheck server"; }
+	return SASL_FAIL;
+    }
+
+    iov[0].iov_base = (char *)userid;
+    iov[0].iov_len = strlen(userid)+1;
+    iov[1].iov_base = (char *)passwd;
+    iov[1].iov_len = strlen(passwd)+1;
+
+    retry_writev(s, iov, 2);
+
+    start = 0;
+    while (start < sizeof(response) - 1) {
+	n = read(s, response+start, sizeof(response) - 1 - start);
+	if (n < 1) break;
+	start += n;
+    }
+
+    close(s);
+
+    if (start > 1 && !strncmp(response, "OK", 2)) {
+	return SASL_OK;
+    }
+
+    response[start] = '\0';
+    if (reply) { *reply = response; }
+    return SASL_BADAUTH;
+}
+
+#endif
