@@ -22,10 +22,20 @@ WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 SOFTWARE.
 ******************************************************************/
-
 #include <config.h>
+#ifndef WIN32
 #include <termios.h>
 #include <unistd.h>
+#else
+#include <stdio.h>
+#include <io.h>
+typedef int ssize_t;
+#define STDIN_FILENO stdin
+#include <saslutil.h>
+__declspec(dllimport) char *optarg;
+__declspec(dllimport) int optind;
+__declspec(dllimport) int getsubopt(char **optionp, char * const *tokens, char **valuep);
+#endif /*WIN32*/
 #include <sasl.h>
 
 #define PW_BUF_SIZE 2048
@@ -34,7 +44,10 @@ static const char build_ident[] = "$Build: saslpasswd " PACKAGE "-" VERSION " $"
 
 const char *progname = NULL;
 
+#ifndef WIN32
+/*doesn't seem to be used anywhere*/
 extern int _sasl_debug;
+#endif /*WIN32*/
 
 void read_password(const char *prompt,
 		   int flag_pipe,
@@ -42,37 +55,76 @@ void read_password(const char *prompt,
 		   unsigned *passlen)
 {
   char buf[PW_BUF_SIZE];
+#ifndef WIN32
   struct termios ts, nts;
   ssize_t n_read;
+#else
+  HANDLE hStdin;
+  DWORD n_read, fdwMode, fdwOldMode;
+  hStdin = GetStdHandle(STD_INPUT_HANDLE);
+  if (hStdin == INVALID_HANDLE_VALUE) {
+	  perror(progname);
+	  exit(-SASL_FAIL);
+  }
+#endif /*WIN32*/
 
   if (! flag_pipe) {
     fputs(prompt, stdout);
     fflush(stdout);
+#ifndef WIN32
     tcgetattr(STDIN_FILENO, &ts);
     nts = ts;
     nts.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHOCTL| ECHOPRT | ECHOKE);
     nts.c_lflag |= ICANON | ECHONL;
     tcsetattr(STDIN_FILENO, TCSAFLUSH, &nts);
+#else
+  if (! GetConsoleMode(hStdin, &fdwOldMode)) {
+	  perror(progname);
+	  exit(-SASL_FAIL);
+  }
+  fdwMode = fdwOldMode & ~ENABLE_ECHO_INPUT;
+  if (! SetConsoleMode(hStdin, fdwMode)) {
+	  perror(progname);
+	  exit(-SASL_FAIL);
+  }
+#endif /*WIN32*/
   }
 
+#ifndef WIN32
   n_read = read(STDIN_FILENO, buf, PW_BUF_SIZE);
   if (n_read < 0) {
+#else
+  if (! ReadFile(hStdin, buf, PW_BUF_SIZE, &n_read, NULL)) {
+#endif /*WIN32*/
+
     perror(progname);
     exit(-SASL_FAIL);
   }
 
   if (! flag_pipe) {
+#ifndef WIN32
     tcsetattr(STDIN_FILENO, TCSANOW, &ts);
     if (0 < n_read && buf[n_read - 1] != '\n') {
       /* if we didn't end with a \n, echo one */
       putchar('\n');
       fflush(stdout);
     }
+#else
+	SetConsoleMode(hStdin, fdwOldMode);
+    putchar('\n');
+    fflush(stdout);
+#endif /*WIN32*/
   }
 
   if (0 < n_read && buf[n_read - 1] == '\n') /* if we ended with a \n */
     n_read--;			             /* remove it */
 
+#ifdef WIN32
+  /*WIN32 will have a CR in the buffer also*/
+  if (0 < n_read && buf[n_read - 1] == '\r') /* if we ended with a \r */
+    n_read--;			             /* remove it */
+#endif /*WIN32*/
+
   *password = malloc(n_read + 1);
   if (! *password) {
     perror(progname);
@@ -108,7 +160,9 @@ main(int argc, char *argv[])
   sasl_conn_t *conn;
   char *user_domain = NULL;
 
+#ifndef WIN32
   _sasl_debug=1;
+#endif /*WIN32*/
 
   if (! argv[0])
     progname = "saslpasswd";
@@ -172,10 +226,12 @@ main(int argc, char *argv[])
 			   &conn);
   if (result != SASL_OK)
     exit_sasl(result, NULL);
-  
+ 
+#ifndef WIN32
   if (! flag_pipe && ! isatty(STDIN_FILENO))
     flag_pipe = 1;
-  
+#endif /*WIN32*/
+
   read_password("Password: ", flag_pipe, &password, &passlen);
 
   if (! flag_pipe) {
